#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import os

from pis.config import (
    config_user_path,
    latest_config_url,
    with_latest_config,
    write_config,
)
from pis import __version__
from pis import lib


def main():
    #TODO:: these lines are module cli, and here is's imported only
    parser = argparse.ArgumentParser(
        prog='pis',
        description=(
            'Install python package as cloned repo from guessed VCS'
            ' (git, hg, etc.) repository.'
        ),
    )
    #TODO::
    #parser.add_argument(
    #    '-d', '--destiny-dir', action='store', default=os.getcwd(),
    #)
    parser.add_argument(
        '-v', '--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
    )
    parser.add_argument(
        'pkg_names', nargs='+',
    )
    args = parser.parse_args()
    if args.pkg_names:
        config = with_latest_config(latest_config_url, config_user_path)
        #founds = lib.source_install(args.pkg_names, args.destiny_dir, config)
        founds = lib.source_install(args.pkg_names, os.getcwd(), config)
        if founds:
            for pkg_name, repo_url in founds.items():
                config['pkg_name2repo_url'][pkg_name] = repo_url
                write_config(config_user_path, config)


if __name__ == "__main__":
    main()



import collections
import json
import logging
import os
import sys

from pip._vendor import requests


try:
    from json.decoder import JSONDecodeError
except ImportError:
    # python < 3.5
    JSONDecodeError = ValueError


log = logging.getLogger(__file__)

#TODO:: rm config_  prefix
config_encoding = 'utf8'
config_filename = 'config.json'
config_default_path = os.path.join(
    os.path.dirname(__file__), config_filename
)
config_user_dir = os.path.join(os.path.expanduser('~'), '.' + __package__)
config_user_path = os.path.join(config_user_dir, config_filename)

latest_config_url = 'https://raw.githubusercontent.com/xliiv/pis/master/pis/config.json'


def update_dikt(dikt_to_update, dikt2):
    for k, v in dikt2.items():
        if (
            k in dikt_to_update and
            isinstance(dikt_to_update[k], dict) and
            isinstance(dikt2[k], collections.Mapping)
        ):
            update_dikt(dikt_to_update[k], dikt2[k])
        else:
            dikt_to_update[k] = dikt2[k]


def init_user_config(config_dir, config_name, config_default):
    config_user_path = os.path.join(config_dir, config_name)
    try:
        read_config(config_user_path)
    except FileNotFoundError:
        try:
            os.mkdir(config_dir)
        except FileExistsError:
            pass
        write_config(config_user_path, config_default)
    except JSONDecodeError:
        log.error(
            "Config file ({}) includes invalid JSON".format(config_user_path)
        )
        sys.exit(1)


def read_config(path):
    #TODO:: merge apps with user config?
    with open(path) as config_file:
        return json.loads(config_file.read())


def write_config(path, config_dict, encoding=config_encoding):
    with open(path, 'wb') as config_file:
        config_file.write(json.dumps(config_dict).encode(encoding))


def download_latest_config(latest_config_url):
    response = requests.get(latest_config_url)
    #TODO: error handling
    assert response.status_code == 200, "Couldn't get latest json"
    return json.loads(response.content.decode())


def with_latest_config(latest_config_url, config_path):
    user_config = read_config(config_path)
    latest_config = download_latest_config(latest_config_url)
    update_dikt(latest_config, user_config)
    return latest_config


init_user_config(
    config_user_dir, config_filename, read_config(config_default_path)
)



import json
import os
import tempfile
import unittest
from unittest import mock

from pis.config import (
    config_default_path,
    latest_config_url,
    init_user_config,
    read_config,
    with_latest_config,
    write_config,
)
from pis.lib import (
    _get_links_from_pypi,
    cd,
    clone_repo,
    source2pkg_name,
    url2vcses,
    verify_pkg_dir,
)


#TODO:: mv to seperate file
SETUPPY_NAME_AS_STRING = '''
from setuptools import setup
setup(
    name="name_as_string",
)
'''

SETUPPY_NAME_AS_VAR = '''
from setuptools import setup
__name__ = 'name_as_dunder_name'

setup(
    name=__name__,
)
'''
SETUPPY_SETUP_AS_ATTRIBUTE = '''
import setuptools

setuptools.setup(
    name="setup_as_attribute",
)
'''


def _config():
    return read_config(config_default_path)


class TestCd(unittest.TestCase):
    def test_cd_works_ok(self):
        tmp_dir = tempfile.TemporaryDirectory()
        init_path = os.getcwd()
        with cd(tmp_dir.name):
            self.assertEqual(os.getcwd(), tmp_dir.name)
        self.assertEqual(init_path, init_path)


class TestUrl2VCSes(unittest.TestCase):
    def test_github_url_returns_url_returns_ok(self):
        url = "http://github.com/xliiv/pis"
        vcs = url2vcses(url, _config())
        self.assertEqual(len(vcs), 1)
        self.assertEqual(vcs[0].name, 'git')

    def test_bitbucket_url_returns_url_returns_ok(self):
        url = "http://bitbucket.org/xliiv/pis"
        vcs = url2vcses(url, _config())
        self.assertEqual(len(vcs), 2)
        self.assertEqual([v.name for v in vcs], ['git', 'hg'])


class TestGetLinksFromPypi(unittest.TestCase):
    def setUp(self):
        class ResponseStub:
            encoding = 'utf-8'
            content = ''.encode('utf-8')
        self.response = ResponseStub()
        self.url = 'https://pypi.python.org/pypi/pis'

    def test_url_found_when_plain_link(self):
        self.response.content = (
            " {} ".format(
                self.url,
            ).encode('utf-8')
        )
        found = _get_links_from_pypi(self.response)
        self.assertEqual([self.url], found)

    def test_url_found_when_link_in_html(self):
        self.response.content = (
            "<a href=\"{}\">naked boobs here!!</a> ".format(
                self.url,
            ).encode('utf-8')
        )
        found = _get_links_from_pypi(self.response)
        self.assertEqual([self.url], found)


class Source2PkgName(unittest.TestCase):
    def test_pkg_name_set_as_string(self):
        pkg_name = source2pkg_name(SETUPPY_NAME_AS_STRING)
        self.assertEqual(pkg_name, 'name_as_string')

    @unittest.skip(
        "needs more details, cos it can't be reproduced with current data"
    )
    def test_pkg_name_set_as_attribute(self):
        pkg_name = source2pkg_name(SETUPPY_NAME_AS_VAR)
        self.assertEqual(pkg_name, 'name_as_dunder_name')

    @unittest.skip("TODO")
    def test_pkg_name_set_when_full_path_import(self):
        #TODO:: this fails when setup.py uses setuptools.setup (instead of "setup(..)")
        pkg_name = source2pkg_name(SETUPPY_SETUP_AS_ATTRIBUTE)
        self.assertEqual(pkg_name, 'setup_as_attribute')


class TestVerifyPkgDir(unittest.TestCase):
    def _pkg_dir(self, pkg_name):
        return os.path.join(
            os.path.dirname(__file__), '_pkg_examples', pkg_name,
        )

    def test_return_true_when_name_matched(self):
        self.assertTrue(verify_pkg_dir(self._pkg_dir('ok_pkg'), 'ok_pkg'))

    def test_return_false_when_name_different(self):
        self.assertFalse(
            verify_pkg_dir(self._pkg_dir('other_pkg'), 'ok_pkg')
        )


class TestCloneRepo(unittest.TestCase):
    @mock.patch('pis.lib._clone_repo')
    def test_clone_repo_returns_true_when_cloned(self, mocked_fn):
        mocked_fn.return_value = True
        result = clone_repo('http://github.com/xliiv/pis', '.', _config())
        self.assertTrue(result)

    @mock.patch('pis.lib._clone_repo')
    def test_clone_repo_returns_false_when_failed(self, mocked_fn):
        mocked_fn.return_value = False
        result = clone_repo('http://github.com/xliiv/pis', '.', _config())
        self.assertFalse(result)


class TestConfigInit(unittest.TestCase):
    def setUp(self):
        tmp = tempfile.TemporaryDirectory()
        self.absent_dir_name = tmp.name
        self.existing_dir = tempfile.TemporaryDirectory()
        self.config_name = 'user_conf.json'
        self.config_default = {
            'var': 'val'
        }

    def test_default_config_is_valid_json(self):
        read_config(config_default_path)

    def test_user_dir_is_created(self):
        init_user_config(
            self.absent_dir_name, self.config_name, self.config_default
        )
        self.assertTrue(os.path.exists(self.absent_dir_name))

    def test_user_config_is_created_when_config_dir_exists(self):
        init_user_config(
            self.existing_dir.name, self.config_name, self.config_default
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(self.existing_dir.name, self.config_name)
            )
        )

    def test_user_config_is_created_when_config_dir_is_absent(self):
        init_user_config(
            self.absent_dir_name, self.config_name, self.config_default
        )
        self.assertTrue(
            os.path.exists(
                os.path.join(self.absent_dir_name, self.config_name)
            )
        )

    def test_created_user_config_is_as_default(self):
        init_user_config(
            self.absent_dir_name, self.config_name, self.config_default
        )
        with open(
            os.path.join(self.absent_dir_name, self.config_name)
        ) as conf:
            content = json.loads(conf.read())
        self.assertEqual(content, self.config_default)

    def test_stop_when_user_config_is_invalid_json(self):
        user_config = tempfile.NamedTemporaryFile()
        user_config.write('invalidjson'.encode('utf8'))
        with self.assertRaises(SystemExit):
            init_user_config(
                os.path.dirname(user_config.name),
                os.path.basename(user_config.name),
                {}
            )


class TestWithLatestConfig(unittest.TestCase):
    def setUp(self):
        self.user_config = tempfile.NamedTemporaryFile()
        write_config(
            self.user_config.name,
            {
                'user': 'user',
                'both': 'user',
                'nested': {
                    'user': 'user',
                    'both': 'user',
                }
            }
        )
        self.latest_config = {
            'latest': 'latest',
            'both': 'latest',
            'nested': {
                'latest': 'latest',
                'both': 'latest',
            }
        }

    @mock.patch('pis.config.download_latest_config')
    def test_latest_value_is_used_when_users_missing(self, latest_config):
        latest_config.return_value = self.latest_config
        config = with_latest_config(latest_config_url, self.user_config.name)
        self.assertEqual(config['latest'], 'latest')

    @mock.patch('pis.config.download_latest_config')
    def test_user_value_is_used_when_latest_missing(self, latest_config):
        latest_config.return_value = self.latest_config
        config = with_latest_config(latest_config_url, self.user_config.name)
        self.assertEqual(config['user'], 'user')

    @mock.patch('pis.config.download_latest_config')
    def test_user_value_is_used_when_both_exists(self, latest_config):
        latest_config.return_value = self.latest_config
        config = with_latest_config(latest_config_url, self.user_config.name)
        self.assertEqual(config['both'], 'user')

    @mock.patch('pis.config.download_latest_config')
    def test_nested_latest_value_is_used_when_users_missing(
        self, latest_config
    ):
        latest_config.return_value = self.latest_config
        config = with_latest_config(latest_config_url, self.user_config.name)
        self.assertEqual(config['nested']['latest'], 'latest')

    @mock.patch('pis.config.download_latest_config')
    def test_nested_user_value_is_used_when_latest_missing(
        self, latest_config
    ):
        latest_config.return_value = self.latest_config
        config = with_latest_config(latest_config_url, self.user_config.name)
        self.assertEqual(config['nested']['user'], 'user')

    @mock.patch('pis.config.download_latest_config')
    def test_nested_user_value_is_used_when_both_exists(self, latest_config):
        latest_config.return_value = self.latest_config
        config = with_latest_config(latest_config_url, self.user_config.name)
        self.assertEqual(config['nested']['both'], 'user')


@unittest.skip("TODO")
class TestSourceInstall(unittest.TestCase):
    def test_install_single_pkg_ok(self):
        import subprocess as subp
        path = '/home/xliiv/workspace/pis/pis/ok.sh'
        path2 = '/home/xliiv/workspace/pis/pis/bad.sh'
        print(path, path2)
        subp.check_call(['bash', 'tmp_dirpath', 'pis_execpath'])

    def test_install_many_pkg_ok(self):
        # source_install('pis', dir_path)
        pass

    def test_install_works_when_dir_passed(self):
        pass

    def test_install_raise_exception_when_dir_not_exists(self):
        pass



import ast
import contextlib
import glob
import logging
import os
import pip
import re
import shutil
import subprocess
import sys
from urllib.parse import urlparse

from pip._vendor import requests
from pip.utils import call_subprocess
from pip.vcs import vcs



logger = logging.getLogger(__name__)

#TODO: user info
#TODO: release with readme, with examples of use, and rm todos from setup.py etc.
#TODO: cli arg dst_path is repo itself this is not what i meant, fix it
#TODO: respect versions by tags or branches
#TODO: docstrings
#TODO: resolve all TODOs (including those from test file)
#TODO: got installed flask=master, then install flask==0.9, it could wipe your whole dir
#       - stop running when dir is found


def _clone_repo(vcs, dst_path):
    try:
        print(vcs.obtain(dst_path))
    except pip.exceptions.InstallationError:
        result = False
    else:
        result = True
    return result


def clone_repo(repo_url, dst_path, config):
    vcses = url2vcses(repo_url, config)
    for VCSClass in vcses:
        full_url = '+'.join([VCSClass.name, repo_url])
        vcs = VCSClass(full_url)
        if _clone_repo(vcs, dst_path):
            return True
        else:
            continue
    return False


@contextlib.contextmanager
def cd(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def url2vcses(url, config):
    vcses = []
    parsed = urlparse(url)
    vcs_names = config['repo_hosts2vcses'].get(parsed.netloc, [])
    for name in vcs_names:
        for registered_vcs in vcs.backends:
            if registered_vcs.name == name:
                vcses.append(registered_vcs)
    return vcses


def source2pkg_name(setuppy_source):
    tree = ast.parse(setuppy_source)
    fl = FuncLister()
    fl.visit(tree)
    return fl.pkg_name


class FuncLister(ast.NodeVisitor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pkg_name = None

    def visit_Call(self, node):
        #TODO:: this failes when setup.py uses setuptools.setup (instead of "setup(..)")
        if getattr(node.func, 'id', '') == 'setup':
            for keyword in node.keywords:
                if keyword.arg == 'name':
                    if isinstance(keyword.value, ast.Str):
                        # handle case: name='pkg_name'
                        self.pkg_name = keyword.value.s
                    elif (
                        # handle case: name=__name__
                        isinstance(keyword.value, ast.Attribute) and
                        keyword.value.attr == '__name__'
                    ):
                        self.pkg_name = keyword.value.value.id
                    else:
                        self.pkg_name = None
        self.generic_visit(node)


def verify_pkg_dir(pkg_dir, pkg_name):
    setup_py_path = os.path.join(pkg_dir, 'setup.py')
    try:
        with open(setup_py_path) as f:
            setuppy_source = f.read()
    except FileNotFoundError:
        is_pkg_dir = False
    else:
        cloned_pkg_name = source2pkg_name(setuppy_source)
        # pypi ignores case, so `pis` and `PIS` is the same package
        is_pkg_dir = cloned_pkg_name.lower() == pkg_name.lower()
    return is_pkg_dir


def _get_links_from_pypi(response, default_encoding='utf-8'):
    content = response.content.decode(response.encoding or default_encoding)
    matcher = re.compile('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')  # noqa
    return matcher.findall(content)


def get_links_from_pypi(pkg_homepage_url, default_encoding):
    """
    Get all links from `pkg_homepage`.
    """
    response = requests.get(pkg_homepage_url)
    return _get_links_from_pypi(response, default_encoding)


def install_pkg(pkg_dir):
    subprocess.check_call([
        "pip{}".format(sys.version[:1]), "install", "-e", pkg_dir,
    ])


#
# DIRT STUFF BEGINS
#

def filter_urls(urls, ok_values):
    #TODO:: write tests
    new_urls = []
    for url in urls:
        for netloc in ok_values:
            if netloc in url:
                new_urls.append(url)
    return new_urls


def source_install(pkges_list, dir_path, config):
    founds = {}
    for pkg_name in pkges_list:
        url = config['pkg_name2repo_url'].get(pkg_name, None)
        if url:
            urls = [url]
        else:
            pkg_homepage = config['pypi_url'].format(pkg_name=pkg_name)
            urls = get_links_from_pypi(pkg_homepage, 'utf-8')
            urls = filter_urls(urls, config['repo_hosts2vcses'].keys())
        for url in urls:
            if clone_repo(repo_url=url, dst_path=pkg_name, config=config):
                if verify_pkg_dir(pkg_dir=pkg_name, pkg_name=pkg_name):
                    install_pkg(pkg_dir=pkg_name)
                    founds[pkg_name] = url
                    break
                else:
                    shutil.rmtree(pkg_name)
    return founds



import setuptools
import sys

from pis import __version__


if int(sys.version[0]) < 3:
    print("Only Python 3.x supported")
    sys.exit(1)


setuptools.setup(
    name="pis",
    version=__version__,
    url="http://github.com/xliiv/pis",

    author="xliiv",
    author_email="tymoteusz.jankowski@gmail.com",

    description=(
        'Install python package as cloned repo from guessed VCS'
        ' (git, hg, etc.) repository.'
    ),
    long_description=open('README.md').read(),
    license='MIT',

    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[],

    entry_points={
        'console_scripts': [
            'pis = pis.main:main',
        ],
    },

    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)
